"""Fs module."""
from .local_file_system import FileInfo, list_path
from .hierarchy import Hierarchy
from .exception import FileInfoNotReasonable

__all__ = [
    "FileInfo",
    "list_path",
    "Hierarchy",
    "FileInfoNotReasonable"
]



"""Hierarchy of parquet file system."""
from enum import Enum


class Hierarchy(Enum):
    """Hierarchy enum of parquet file system."""

    REPO, SET, HADOOP_SET, HADDOP_PARTITION, FILE = range(5)

    def descend(self, directory: bool=False,
                hadoop: bool=False) -> Enum:
        """Descend hierarchy.

        Args:
            hierarchy (Hierarchy): ancestor hierarchy
            directory (bool): is dir?
            haddop (bool): is hadoop?
        """
        # not directory, then must be file
        if not directory:
            return Hierarchy.FILE

        hierarchy = self.value
        if hierarchy == Hierarchy.REPO.value:
            if hadoop:
                return Hierarchy.HADOOP_SET
            return Hierarchy.SET
        if hierarchy == Hierarchy.HADOOP_SET.value:
            return Hierarchy.HADDOP_PARTITION
        return Hierarchy.FILE



"""Path utils."""
import os


def dir_to_name(dir_path: str) -> str:
    """full file path to a simple name"""
    _split = dir_path.split(os.sep)
    _full_name = _split[len(_split) - 1]
    _dot_index = _full_name.find(".")
    if _dot_index > -1:
        _full_name = _full_name[0: _dot_index]
    return _full_name



"""Paruqet metadata loader."""
from pyarrow import dataset as ds

from blastoise.fs import FileInfo, Hierarchy
from blastoise.util import dir_to_name
from .exception import RepoDirectoryCantLoadAloneEception


class ParquetLoader:
    """Load metadata of paruqet files of the same struct in memory."""

    def __init__(self, file_info: FileInfo) -> None:
        """Constructor."""
        if file_info._hierarchy == Hierarchy.REPO:
            raise RepoDirectoryCantLoadAloneEception()

        # file or directory name to dataset name
        _path = file_info.name
        self.path = _path
        self.ds_name = dir_to_name(_path)
        self.size = file_info.size
        self.hierarchy = file_info._hierarchy
        self.directory = file_info._directory
        self.hadoop = file_info._hadoop
        self.dataset = ParquetLoader.load_dataset(file_info)

    @classmethod
    def load_dataset(cls, file_info: FileInfo):
        """Load file to dataset."""
        hier = file_info.hierarchy()
        if hier == Hierarchy.REPO:
            raise RepoDirectoryCantLoadAloneEception()

        if hier == Hierarchy.SET:
            return [ds.dataset([info.name for info in file_group], format="parquet") \
                    for file_group in file_info.spllit_dir()]

        return [ds.dataset(file_info.name, format="parquet")]

    @classmethod
    def load(cls, file_info: FileInfo | list) -> list:
        """Load any parquet path to dataset."""
        if isinstance(file_info, list):
            res = []
            for file in file_info:
                res += cls.load(file)
            return res
        if file_info.hierarchy() == Hierarchy.REPO:
            return [ParquetLoader(f) for f in file_info.list_dir()]
        return [ParquetLoader(file_info)]



# -*- coding: utf-8 -*-
try:
    from setuptools import find_packages, setup
except ImportError:
    import distribute_setup

    distribute_setup.use_setuptools()
    from setuptools import setup, find_packages

import os
import sys
from distutils import log

long_desc = """Blastoise is a python tool lib that helps reading and querying data from parquet files. After querying, you can stroe it in Plasma memory map and do some futher calculations."""

requires = ["pyarrow", "pandas"]

setup(
    name="blastoise",
    version="0.0.1",
    url="https://github.com/Hebrh/blastoise",
    license="MIT License",
    author="zhujw",
    author_email="kratoswittgenstein@gmail.com",
    description="a stream tool that helps you read and query data from parquet files",
    long_description=long_desc,
    zip_safe=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=find_packages(),
    include_package_data=True,
    install_requires=requires,
)



"""Utils for local file system."""
from pyarrow import fs
from pyarrow.fs import FileType

from blastoise import MAX_FILES_SIZE
from .exception import FileInfoNotReasonable
from .hierarchy import Hierarchy


class FileInfo():
    """Simple file info data."""

    # pylint: disable = too-many-arguments
    def __init__(
        self, name: str, size: int=0,
        hierarchy: Hierarchy=Hierarchy.REPO,
        directory: bool=False,
        hadoop: bool=False
        ) -> None:
        """Constructor for FileInfo.

        Args:
            name (str): file or directory path
            hierarchy (Hierarchy): hierarchy
            directory (bool): true if it's directory
            hadoop (bool): is hadoop file system
        """
        if hierarchy == Hierarchy.FILE and directory:
            raise FileInfoNotReasonable()

        # adjust path not to end with slash
        if name.endswith("/"):
            name = name[0: len(name) - 1]
        self.name = name
        if size == 0:
            self.size = 0
        else:
            self.size = size / (1024 * 1024 * 1.0)
        self._hierarchy = hierarchy
        self._directory = directory
        self._hadoop = hadoop
        self._children = []

    def is_dir(self) -> bool:
        """Whether it's directory."""
        return self._directory

    def hierarchy(self) -> Hierarchy:
        """Getter for hierarchy."""
        return self._hierarchy

    def descend_hierarchy(self, directory: bool) -> Hierarchy:
        """Descend hierarchy."""
        return self._hierarchy.descend(directory=directory, hadoop=self._hadoop)

    def add_child(self, name: str, size: int, directory: bool) -> None:
        """Add child for directory.

        Args:
            name (str): file or directory path
            directory (bool): true if it's directory
        """
        # pass if not a directory
        if not self._directory:
            return

        child_count = len(self._children)
        if child_count > 0:
            tail_child = self._children[child_count - 1]
            # tail add child
            if tail_child.is_dir() and name.startswith(tail_child.name + "/"):
                tail_child.add_child(name, size, directory)
                return

        self._children.append(FileInfo(name,  size, hierarchy=self.descend_hierarchy(directory),
                    directory=directory))

    def list_dir(self) -> list:
        """List the directory.
        Return empty list for file path.
        """
        return self._children

    def spllit_dir(self) -> list:
        """Split children file list by MAX_FILES_SIZE."""
        if self._hierarchy == Hierarchy.SET:
            max_size = MAX_FILES_SIZE
            files = self._children
            split_files = []
            file_group = []
            total_size = 0
            total_more_size = 0

            i = 0
            f_len = len(files)
            while i < f_len:
                file = files[i]
                total_more_size += total_size + file.size
                if total_more_size <= max_size:
                    file_group.append(file)
                    total_size = total_more_size
                    i += 1
                    continue
                if total_more_size - max_size < max_size - total_size:
                    file_group.append(file)
                    i += 1
                split_files.append(file_group)
                file_group = []
                total_size = 0
                total_more_size = 0

        return split_files

    def __repr__(self) -> str:
        return f"FileInfo:[path={self.name}, id_dir={self._directory}, hier={self._hierarchy}," \
            f"size={self.size}, hadoop={self._hadoop}, child={self._children}]"


def list_path(path: str) -> list:
    """List file paths in a parent directory.
    or empty list if 'path' not exists.

    Args:
        path (str): a file path
    """
    local = fs.LocalFileSystem()
    selector = fs.FileSelector(path, allow_not_found=True, recursive=True)
    file_infos = local.get_file_info(selector)
    parent_directory = FileInfo(path, directory=True)
    for file_info in file_infos:
        is_dir = file_info.type == FileType.Directory
        if is_dir:
            size = 0
        else:
            size = file_info.size
        parent_directory.add_child(file_info.path, size, is_dir)
    return parent_directory.list_dir()
